/*
 *  Copyright (C) 2005-2006 mark307
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

extern "C" {
#include <lua.h>
#include <lauxlib.h>
}

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <kmeleon_plugin.h>

#include "macroset.h"

extern kmeleonPlugin kPlugin;
extern kmeleonFunctions *kFuncs;
extern MacroSet* g_lpMacroset;
extern CRITICAL_SECTION g_hCriticalSection;

MacroSet::~MacroSet()
{
	Node *p;
	int i;
	
	for (i = 0, p = node; i < length; i ++, p ++) {
		free(p->macro);
		free(p->arg);
	}
	free(node);
}


int MacroSet::add(const char *macro, const char *arg, int ref)
{
	Node *p = NULL;
	char *m = NULL, *a = NULL;
	
	p = (Node *) realloc(node, sizeof(Node) * (length + 1));
	m = _strdup(macro);
	a = _strdup(arg);
	if (! p || ! m || ! a) {
		free(p); free(m); free(a);
		return -1;
	}
	node = p;
	p = node + length;
	length ++;
	p->id = kFuncs->GetCommandIDs(1);
	p->macro = m;
	p->arg = a;
	p->ref = ref;
	return p->id;
}


int MacroSet::replace(const char *macro, const char *arg, int *ref)
{
	Node *p;
	int i;
	
	for (i = 0, p = node; i < length; i ++, p ++) {
		if (! strcmp(macro, p->macro) && ! strcmp(arg, p->arg)) {
			i = p->ref;
			p->ref = *ref;
			*ref = i;
			return p->id;
		}
	}
	return add(macro, arg, *ref);
}


const int MacroSet::search(const char *macro, const char *arg) const
{
	Node *p;
	int i;
	
	for (i = 0, p = node; i < length; i ++, p ++) {
		if (! strcmp(macro, p->macro) && ! strcmp(arg, p->arg)) {
			return p->id;
		}
	}
	return -1;
}


const MacroSet::Node* MacroSet::search(int id) const
{
	int min = 0, max = length - 1;
	int i = -1;
	
	if (! node || id < node[min].id || id > node[max].id) {
		return NULL;
	}
	while (min < max) {
		i = (min + max) / 2;
		if (node[i].id < id) {
			min = i + 1;
		} else {
			max = i;
		}
	}
	if (node[min].id == id) {
		return &node[min];
	}
	return NULL;
}


int parse_plugin_param(char *param, char **macro, char **arg, char **str)
{
	char *p, *me = NULL;
	int n = 0;
	
	*macro = *arg = *str = NULL;
	p = param;
	// avoid space prefix
	while (isspace(*p)) p ++;
	if (*p == '\0') {
		return n;
	}
	// begin identifier
	*macro = p;
	n ++;
	// determine end of identifier
	while (isalnum(*p) || *p == '_' || *p == '.') p ++;
	if (*p == '\0') {
		return n;
	}
	// end identifier
	me = p;
	// avoid space between identifier and parenthesis
	while (isspace(*p)) p ++;
	// determine end of argument
	if (*p == '(') {
		*arg = ++ p;
		while (*p != ')') {
			if (*p == '\0') {
				*me = '\0';
				return n;
			}
			p ++;
		}
		*p ++ = '\0';
		n ++;
		while (isspace(*p)) p ++;
	}
	if (*p == ',') {
		p ++;
		while (isspace(*p)) p ++;
		*str = p;
		n ++;
	}
	*me = '\0';
	return n;
}


int get_function_ref(lua_State *L)
{
	GetRefParam *par = (GetRefParam *) lua_touserdata(L, 1);
	int ref = LUA_NOREF;
	const char *macro = par->macro;
	const char *p;
	
	par->ref = LUA_NOREF;
	if (p = strchr(macro, '.')) {
		lua_pushlstring(L, macro, p - macro);
		lua_gettable(L, LUA_GLOBALSINDEX);
		if (lua_isnil(L, -1)) {
			return 0;
		}
		macro = p + 1;
		while (p = strchr(macro, '.')) {
			lua_pushlstring(L, macro, p - macro);
			lua_gettable(L, -2);
			lua_remove(L, -2);
			if (lua_isnil(L, -1)) {
				return 0;
			}
			macro = p + 1;
		}
		lua_pushstring(L, macro);
		lua_gettable(L, -2);
		lua_remove(L, -2);
	} else {
		lua_pushstring(L, macro);
		lua_gettable(L, LUA_GLOBALSINDEX);
	}
	if (lua_isfunction(L, -1)) {
		ref = lua_ref(L, 1);
	}
#if 0
	/* little tricky
	   evalulate 'do return "macro name with table" end' */
	lua_pushstring(L, "return ");
	lua_pushstring(L, macro);
	lua_concat(L, 2);
	if (! luaL_loadbuffer(L, lua_tostring(L, -1), lua_strlen(L, -1), NULL)) {
		if (! lua_pcall(L, 0, 1, 0)) {
			if (lua_isfunction(L, -1)) {
				ref = lua_ref(L, 1);
			}
		}
	}
#endif
	par->ref = ref;
	return 0;
}
