/*
 * Copyright (C) 2005-2006 mark307
 * See Copyright Notice in COPYING.LUA
 */

#define WIN32_LEAN_AND_MEAN
//#define _WIN32_WINNT 0x0403
#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <stdlib.h>

extern "C" {
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
}

#include "luamacros.h"
#include "kmenv.h"
#include "luaut.h"
#include "w32.h"

/**
   
   * Windows functions
   
   ** Defined variables
   
   ICON_EXCLAIM ICON_INFO ICON_STOP ICON_QUESTION
   BUTTONS_OK BUTTONS_RETRYCANCEL BUTTONS_YESNO
   BUTTONS_YESNOCANCEL BUTTONS_ABORTRETRYIGNORE
   
      Dialog type and decoration, like MessageBox Win32 API.
   
   IDOK IDYES IDNO IDABORT IDRETRY IDIGNORE IDCANCEL
   
      Return value of confirm, like MessageBox Win32 API.
   
   
   ** Defined functions
   
 */
const lut_IntReg w32_consts[] = {
	/*
	 * message box
	 */
	{"ICON_EXCLAIM", MB_ICONEXCLAMATION},
	{"ICON_INFO", MB_ICONINFORMATION},
	{"ICON_STOP", MB_ICONSTOP},
	{"ICON_QUESTION", MB_ICONQUESTION},
	{"BUTTONS_OK", MB_OK},
	{"BUTTONS_RETRYCANCEL", MB_RETRYCANCEL},
	{"BUTTONS_YESNO", MB_YESNO},
	{"BUTTONS_YESNOCANCEL", MB_YESNOCANCEL},
	{"BUTTONS_ABORTRETRYIGNORE", MB_ABORTRETRYIGNORE},
	{"IDOK", IDOK},
	{"IDYES", IDYES},
	{"IDNO", IDNO},
	{"IDABORT", IDABORT},
	{"IDRETRY", IDRETRY},
	{"IDIGNORE", IDIGNORE},
	{"IDCANCEL", IDCANCEL},
	/* end */
	{NULL, 0},
};


/**
   exec(command [, timeout])
   
      Executes an external program.
   
   command : string
   timeout : number : timeout for waiting process termination
   
   A unit of timeout is millisecond.
   
   Timeout must be greater than 0 and less than 10000 (10sec)
   
   return : boolean
 */
int L_exec(lua_State *L)
{
	static const int timeout_max = 10000;
	char *cmd;
	int timeout;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	BOOL ret;
	
	luaL_checkany(L, 1);
	timeout = luaL_optinteger(L, 2, 0);
	cmd = lut_checkstring(L, 1);
	if (! strlen(cmd)) {
		return 0;
	}
	
	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(STARTUPINFO);
	ret = CreateProcess(NULL, cmd, NULL, NULL, FALSE, 0,
						NULL, NULL, &si, &pi);
	if (ret) {
		CloseHandle(pi.hThread);
		if (timeout > 0) {
			WaitForSingleObject(pi.hProcess,
								timeout > timeout_max ? timeout_max : timeout);
		} else if (timeout < 0) {
			WaitForInputIdle(pi.hProcess,
							 timeout < - timeout_max ? timeout_max : - timeout);
		}
		CloseHandle(pi.hProcess);
	}
	lua_pushboolean(L, ret);
	return 1;
}


/**
   alert([message [, title [, icon]]])
   
      Shows a messagebox.
   
   message, title : string
   icon : ICON_EXCLAIM | ICON_INFO | ICON_STOP | ICON_QUESTION
*/
int L_alert(lua_State *L)
{
	const char *message, *title;
	int icon;
	HWND hwnd;
	
	hwnd = get_hwnd(L);
	message = luaL_optstring(L, 1, "alert");
	title   = luaL_optstring(L, 2, "K-Meleon");
	icon    = luaL_optinteger(L, 3, 0);
	MessageBox(hwnd, message, title, icon | MB_OK | MB_TASKMODAL);
	return 0;
}


/**
   confirm([message [, title [, buttons [, icon]]]])
   
      Shows a messagebox and ask for confirmation.
   
   message, title : string
   buttons : BUTTONS_OK | BUTTONS_RETRYCANCEL | BUTTONS_YESNO |
             BUTTONS_YESNOCANCEL | BUTTONS_ABORTRETRYIGNORE
   icon : ICON_EXCLAIM | ICON_INFO | ICON_STOP | ICON_QUESTION
   
   return : IDOK | IDYES | IDNO | IDABORT | IDRETRY | IDIGNORE | IDCANCEL
*/
int L_confirm(lua_State *L)
{
	const char *message, *title;
	int icon, buttons;
	HWND hwnd;
	int ret;
	
	hwnd = get_hwnd(L);
	message = luaL_optstring(L, 1, "confirm");
	title   = luaL_optstring(L, 2, "K-Meleon");
	icon    = luaL_optinteger(L, 3, 0);
	buttons = luaL_optinteger(L, 4, MB_OK);
	ret = MessageBox(hwnd, message, title, buttons | icon | MB_TASKMODAL);
	lua_pushinteger(L, ret);
	return 1;
}


struct PromptDialogParam {
	const char *title;
	const char *prompt;
	const char *default_value;
	char *answer;
};


BOOL CALLBACK
PromptDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static const TCHAR lpParam[] = "param";
	struct PromptDialogParam *param;
	HWND hwndDlgItem;
	int n;
	
	switch (message) {
	case WM_INITDIALOG:
		param = (struct PromptDialogParam *) lParam;
		if (param->title) {
			SetWindowText(hwnd, param->title);
		}
		if (param->prompt) {
			SetDlgItemText(hwnd, IDC_PROMPT, param->prompt);
		}
		if (param->default_value) {
			SetDlgItemText(hwnd, IDC_ANSWER, param->default_value);
		}
		SetProp(hwnd, lpParam, param);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDOK:
			param = (struct PromptDialogParam *) GetProp(hwnd, lpParam);
			hwndDlgItem = GetDlgItem(hwnd, IDC_ANSWER);
			n = GetWindowTextLength(hwndDlgItem) + 1;
			param->answer = (char *) malloc(n);
			if (param->answer) {
				GetDlgItemText(hwnd, IDC_ANSWER, param->answer, n);
			}
			RemoveProp(hwnd, lpParam);
			EndDialog(hwnd, IDOK);
			break;
		case IDCANCEL:
			RemoveProp(hwnd, lpParam);
			EndDialog(hwnd, IDCANCEL);
			break;
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}


/**
   prompt([, question [, title [, default]]])
   
      Shows a modal 'prompt' dialog that returns a string value.
      If user will click "Cancel", this function will return nil.
   
   question, title, default : string
   
   return : string or nil
*/
int L_prompt(lua_State *L)
{
	HWND hwnd;
	HINSTANCE hinstance;
	int ret;
	struct PromptDialogParam param = {NULL};
	
	hwnd = get_hwnd(L);
	hinstance = get_dllinstance(L);
	param.prompt = luaL_optstring(L, 1, NULL);
	param.title = luaL_optstring(L, 2, NULL);
	param.default_value = luaL_optstring(L, 3, NULL);
	
	ret = DialogBoxParam(hinstance,
	                     MAKEINTRESOURCE(IDD_PROMPT), hwnd,
	                     (DLGPROC)PromptDlgProc, (LPARAM) &param);
	PostMessage(hwnd, WM_NULL, 0, 0);
	if (ret == IDOK) {
		lua_pushstring(L, param.answer);
		free(param.answer);
		return 1;
	}
	return 0;
}


struct ListDialogParam {
	const char *title;
	const char *prompt;
	const char **values;
	int answer;
};


BOOL CALLBACK
ListDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static const TCHAR lpParam[] = "param";
	struct ListDialogParam *param;
	HWND hwndDlgItem;
	const char **p;
	
	switch (message) {
	case WM_INITDIALOG:
		param = (struct ListDialogParam *) lParam;
		if (param->title) {
			SetWindowText(hwnd, param->title);
		}
		if (param->prompt) {
			SetDlgItemText(hwnd, IDC_PROMPT, param->prompt);
		}
		hwndDlgItem = GetDlgItem(hwnd, IDC_LIST);
		for (p = param->values; p != NULL && *p != NULL; p ++) {
			ListBox_AddString(hwndDlgItem, *p);
		}
		SetProp(hwnd, lpParam, param);
		return TRUE;
	case WM_COMMAND:
		if (LBN_DBLCLK == HIWORD(wParam) || IDOK == LOWORD(wParam)) {
			param = (struct ListDialogParam *) GetProp(hwnd, lpParam);
			hwndDlgItem = GetDlgItem(hwnd, IDC_LIST);
			param->answer = ListBox_GetCurSel(hwndDlgItem);
			RemoveProp(hwnd, lpParam);
			EndDialog(hwnd, IDOK);
		} else if (IDCANCEL == LOWORD(wParam)) {
			RemoveProp(hwnd, lpParam);
			EndDialog(hwnd, IDCANCEL);
			break;
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}


/**
   choose(question, title, ...)
   select(question, title, ...)
   
      Shows a modal 'select' dialog that returns a index and
      string of selected item. Index starts from 1.
      If user clicks "Cancel", this function returns nil.
   
   question       : string : 
   title          : string : dialog title
   rest arguments : string : select items
   
   return : number, string : index of items, text of item
*/
int L_select(lua_State *L)
{
	HWND hwnd;
	HINSTANCE hinstance;
	int n = lua_gettop(L);
	int i, ret;
	struct ListDialogParam param = {NULL};
	
	if (n < 2) {
		return luaL_argerror(L, n, "too few arguments");
	}
	hwnd = get_hwnd(L);
	hinstance = get_dllinstance(L);
	param.prompt = luaL_checklstring(L, 1, NULL);
	param.title = luaL_checklstring(L, 2, NULL);
	/*
	 * -2 means prompt and title.
	 * +1 means NULL terminate.
	 */
	param.values = (const char **) malloc(sizeof(char *) * (n - 2 + 1));
	if (! param.values) {
		return 0;
	}
	for (i = 3; i <= n; i ++) {
		if (lua_isstring(L, i)) {
			param.values[i - 3] = lua_tostring(L, i);
		} else {
			param.values[i - 3] = "";
		}
	}
	param.values[i - 3] = NULL;
	param.answer = LB_ERR;
	ret = DialogBoxParam(hinstance,
	                     MAKEINTRESOURCE(IDD_LIST), hwnd,
	                     (DLGPROC)ListDlgProc, (LPARAM) &param);
	PostMessage(hwnd, WM_NULL, 0, 0);
	i = 0;
	if (ret == IDOK && param.answer != LB_ERR) {
		lua_pushinteger(L, param.answer + 1);
		lua_pushstring(L, param.values[param.answer]);
		i = 2;
	}
	free(param.values);
	return i;
}


/**
   popupmenu(...)
   
      Shows popup menu and returns a index. Index starts from 1.
      If menu is canceled (user clicks without menu),
      this function returns nil.
   
   hwnd           : number : window handle.
   rest arguments : string : menu items.
   
   return : number : index seledted by user.
*/
int L_popupmenu(lua_State *L)
{
	HMENU menu;
	HWND hwnd;
	int n = lua_gettop(L);
	int i, r = 0;
	
	hwnd = get_hwnd(L);
	menu = CreatePopupMenu();
	if (menu) {
		for (i = 1; i <= n; i ++) {
			if (lua_isstring(L, i)) {
				if (lua_strlen(L, i) > 0) {
					AppendMenu(menu, MF_ENABLED | MF_STRING, 0x7fff - i, lua_tostring(L, i));
				} else {
					AppendMenu(menu, MF_SEPARATOR, 0, NULL);
				}
			}
		}
		if (GetMenuItemCount(menu) > 0) {
			POINT pt = {0};
			GetCursorPos(&pt);
			r = TrackPopupMenu(menu, TPM_TOPALIGN | TPM_LEFTALIGN | TPM_RETURNCMD | TPM_NONOTIFY,
			                   pt.x, pt.y, 0, hwnd, NULL);
		}
		DestroyMenu(menu);
	}
	if (r > 0) {
		lua_pushinteger(L, 0x7fff - r);
		return 1;
	}
	return 0;
}


/**
   getclipboard()
   
      Gets text from the clipboard.
   
   return : string
*/
int L_getclipboard(lua_State *L)
{
	HGLOBAL hglobal;
	char *psz = NULL;
	HWND hwnd;
	
	hwnd = get_hwnd(L);
	
	if (OpenClipboard(hwnd)) {
		hglobal = GetClipboardData(CF_TEXT);
		if (hglobal != NULL) {
			psz = (char *) GlobalLock(hglobal);
			lua_pushstring(L, psz);
			GlobalUnlock(hglobal);
		}
		CloseClipboard();
	}
	if (! psz) {
		lua_pushstring(L, "");
	}
	return 1;
}


/**
   setclipboard(text)
   
      Sets text to the clipboard.
   
   text : string
*/
int L_setclipboard(lua_State *L)
{
	HGLOBAL hglobal;
	const char *str;
	char *psz;
	size_t size;
	HWND hwnd;
	
	luaL_checkany(L, 1);
	str = luaL_checklstring(L, 1, NULL);
	hwnd = get_hwnd(L);
	
	size = strlen(str) + 1;
	hglobal = GlobalAlloc(GHND, sizeof(char) * (size));
	if (hglobal == NULL) {
		return 0;
	}
	psz = (char*) GlobalLock(hglobal);
	if (psz == NULL || ! OpenClipboard(hwnd)) {
		GlobalFree(hglobal);
		return 0;
	}
	strcpy(psz, str);
	GlobalUnlock(hglobal);
	EmptyClipboard();
	SetClipboardData(CF_TEXT, hglobal);
	CloseClipboard();
	return 0;
}


/**
   sendmessage(hwnd, id, wparam, lparam)
   
      Wrapper of SendMessage Win32API.
   
   hwnd   : number : window handle
   id     : number : message id
   wparam : number : param
   lparam : number : param
*/
int L_sendmessage(lua_State *L)
{
	luaL_checkany(L, 4);
	lua_pushinteger(L,
					(lua_Integer) SendMessage(
						(HWND) luaL_checkinteger(L, 1),
						(UINT) luaL_checkinteger(L, 2),
						(WPARAM) luaL_checkinteger(L, 3),
						(LPARAM) luaL_checkinteger(L, 4)));
	return 1;
}

/**
   sendkey(vk, scan, flags, extra, ...)
   
      Wrapper of keybd_event Win32API.
      This also accepts sequence of argument sets.
      
      This code sends "Page Down" once.
      km.sendkey(34, 0, 0, 0, 34, 0, 2, 0)
      
      This code sends "CTRL-N".
      km.sendkey(17, 0, 0, 0, 78, 0, 0, 0, 17, 0, 2, 0, 78, 0, 2, 0)
   
   vk     : number : virtual-key code
   scan   : number : hardware scan code (usually zero)
   flags  : number :
   extra  : number :
*/
int L_sendkey(lua_State *L)
{
	int i, j, n;
	
	n = lua_gettop(L);
	if (n < 1 || n % 4 != 0) return 0;
	n /= 4;
	j = 1;
#if (_WIN32_WINNT >= 0x0403)
	{
		OSVERSIONINFO ovi;
		ovi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		if (! GetVersionEx(&ovi)) return 0;
		if (ovi.dwPlatformId == VER_PLATFORM_WIN32_NT) {
			INPUT *input = (INPUT *) lua_newuserdata(L, sizeof(INPUT) * n);
			for (i = 0; i < n; i ++) {
				input[i].type = INPUT_KEYBOARD;
				input[i].ki.wVk = (WORD) luaL_checkinteger(L, j ++);
				input[i].ki.wScan = (WORD) luaL_checkinteger(L, j ++);
				input[i].ki.dwFlags = (DWORD) luaL_checkinteger(L, j ++);
				input[i].ki.time = 0;
				input[i].ki.dwExtraInfo = (WORD) luaL_checkinteger(L, j ++);
			}
			SendInput(n, input, sizeof(INPUT));
			return 0;
		}
	}
#endif /* (_WIN32_WINNT >= 0x0403) */
	for (i = 0; i < n; i ++, j += 4) {
		keybd_event((BYTE) luaL_checkinteger(L, j),
					(BYTE) luaL_checkinteger(L, j + 1),
					(DWORD) luaL_checkinteger(L, j + 2),
					(DWORD) luaL_checkinteger(L, j + 3));
	}
	return 0;
}

const luaL_Reg w32_funcs[] = {
	/* w32 */
	{"exec", L_exec},
	{"alert", L_alert},
	{"confirm", L_confirm},
	{"prompt", L_prompt},
	{"select", L_select},
	{"choose", L_select},
	{"popupmenu", L_popupmenu},
	{"getclipboard", L_getclipboard},
	{"setclipboard", L_setclipboard},
	{"sendmessage", L_sendmessage},
	{"sendkey", L_sendkey},
	/* end */
	{NULL, 0},
};
